#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <pthread.h>
#include <getopt.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xatom.h>

#include "xgroove.h"
#include "grid.h"
#include "network.h"
#include "xgslave.h"
#include "gx11.h"

#define DEFAULT_SLAVE_UDP_PORT 8503

void
do_gslave (int argc, char *argv[])
{
  GSlave     *gslave;
  int        res;

  printf ("Initializing XGroove v1.0 - Slave mode\n");

  if (NULL == (gslave = init_gslave(argv[2])) ) {
    fprintf (stderr, "do_gslave: failed to init_gslave()\n");
    return;
  }

  if (0 > (res = pthread_create (&gslave->xevent_thread, NULL, (void *)&gslave_xevent_loop, gslave)) ) {
    fprintf (stderr, "init_gslave: xevent loop pthread_create failed, %s\n", strerror(res));
    deinit_gslave (gslave);
    return;
  }

  while (1) {
    gslave_loop (gslave);
    sleep (1);
  }

  deinit_gslave (gslave);
}

void
gslave_loop (GSlave *gslave)
{
  int        res;

  while (gslave->tcp_conn == NULL || gslave->tcp_conn->is_connected != 1) {
    if (0 > gslave_find_master (gslave)) {
      fprintf (stderr, "gslave_loop: failed to find a master...pausing\n");
      sleep (1);
    }
    else {
      break;
    }
  }

  while (gslave->tcp_conn->is_connected == 1) {
    GEvent *gevent = init_gevent ();

    if (0 > (res = tcp_recv (gslave->tcp_conn, gevent)) ) {
      fprintf (stderr, "gslave_loop: it appears the socket to the master closed. bye bye\n");
      tcp_close_connection (gslave->tcp_conn);
      gslave->tcp_conn = NULL;
      deinit_gevent (gevent);

      if (0 != gslave->is_pressed) {
        XTestFakeButtonEvent (gslave->gdisplay.display, gslave->is_pressed, False, 0);
        XUngrabServer (gslave->gdisplay.display);
        XFlush (gslave->gdisplay.display);
        gslave->is_pressed = 0;
      }

      return;
    }

    gslave_proc (gslave, gevent);
    deinit_gevent (gevent);
  }

}

void
gslave_proc (GSlave *gslave, GEvent *gevent)
{
  switch (gevent->generic.type) {
    case GrooveMotionEvent:
      gslave_proc_motion (gslave, gevent);
      break;
    case GrooveClickEvent:
      gslave_proc_click (gslave, gevent);
      break;
    case GrooveKeyPressEvent:
      gslave_proc_keypress (gslave, gevent);
      break;
    case GrooveControlEvent:
      gslave_proc_control (gslave, gevent);
      break;
    case GrooveHelloEvent:
      gslave_proc_hello (gslave, gevent);
      break;
    case GrooveHelloBackEvent:
      gslave_proc_helloback (gslave, gevent);
      break;
    case GrooveGoodbyeEvent:
      gslave_proc_goodbye (gslave, gevent);
      break;
    case GrooveClipboardEvent:
      gslave_proc_clipboard (gslave, gevent->gclipboard);
      break;
    default:
      printf ("gslave_proc: unknown groove event type: %d\n", gevent->generic.type);
  }
}

GSlave *
init_gslave (char *name)
{
  GSlave     *gslave;
  GDisplay   *gdisplay;

  if (NULL == (gslave = malloc (sizeof(struct _GSlave))) ) {
    fprintf (stderr, "init_gslave: failed doing malloc\n");
    return NULL;
  }

  gdisplay = &gslave->gdisplay;

  strcpy (gslave->name, name);
  gslave->udp_conn = udp_open_connection ("", DEFAULT_SLAVE_UDP_PORT);
  gslave->tcp_conn = NULL;
  gslave->has_control = 0;
  gslave->is_pressed = 0;
  gslave->prev_pos_h = 0;
  gslave->prev_pos_v = 0;
  gslave->gdisplay.display = NULL;
  gslave->clipboard = NULL;
  gslave->clipboard_len = 0;
  gslave->is_controlled = 0;
  
  if (NULL == (gslave->gdisplay.display = XOpenDisplay (NULL)) ) {
    fprintf (stderr, "init_gslave: failed to connect to x server at %s\n", getenv("DISPLAY"));
    deinit_gslave (gslave);
    return NULL;
  }

  gdisplay->width = DisplayWidth (gdisplay->display, DefaultScreen (gdisplay->display));
  gdisplay->height = DisplayHeight (gdisplay->display, DefaultScreen (gdisplay->display));
  gdisplay->rootwin = &DefaultRootWindow (gdisplay->display);
  gdisplay->mainwin = XCreateSimpleWindow (gdisplay->display, *gdisplay->rootwin,
                                           100, 100, 100, 100, 2, 1, 1);

  udp_set_sotimeout (gslave->udp_conn, 20000);

  return gslave;
}

void
deinit_gslave (GSlave *gslave)
{
  printf ("deinit_gslave: * * * * *\n");

  if (gslave == NULL)
    return;

  if (gslave->udp_conn != NULL) {
    udp_close_connection (gslave->udp_conn);
    gslave->udp_conn = NULL;
  }

  if (gslave->tcp_conn != NULL) {
    tcp_close_connection (gslave->tcp_conn);
    gslave->tcp_conn = NULL;
  }

  if (gslave->clipboard_len > 0) {
    free (gslave->clipboard);
    gslave->clipboard = NULL;
    gslave->clipboard_len = 0;
  }

  if (gslave->gdisplay.display != NULL)
    XCloseDisplay (gslave->gdisplay.display);

  free (gslave);
}

int
gslave_find_master (GSlave *gslave)
{
  UDPMesg mesg_out, mesg_in;
  int     res;
  GEvent  *gevent_config;
  Display *display;

  display = gslave->gdisplay.display;
  mesg_out.gevent = init_gevent ();
  mesg_in.gevent = init_gevent ();
  gevent_config = init_gevent ();

  mesg_out.gevent->ghello.type = GrooveHelloEvent;
  strncpy (mesg_out.gevent->ghello.name, gslave->name, strlen(gslave->name));
  mesg_out.gevent->ghello.port = DEFAULT_SLAVE_UDP_PORT;

  while (1) {
    printf ("gslave_find_master: Searching for master...\n");
/*  udp_send (gslave->udp_conn, &mesg_out, "255.255.255.255", DEFAULT_UDP_PORT); */
    udp_send (gslave->udp_conn, &mesg_out, "10.0.1.255",      DEFAULT_UDP_PORT);

    if (0 > (res = udp_recv (gslave->udp_conn, &mesg_in)) ) {
      sleep (1);
      continue;
    }

    break;
  }

  printf ("gslave_find_master: received reply from %s:%d\n", mesg_in.addr, mesg_in.gevent->ghelloback.port);

  if (NULL == (gslave->tcp_conn = tcp_open_connection (mesg_in.addr, mesg_in.gevent->ghelloback.port)) ) {
    fprintf (stderr, "gslave_find_master: failed to connect to master\n");
    deinit_gevent (mesg_out.gevent);
    deinit_gevent (mesg_in.gevent);
    deinit_gevent (gevent_config);
    return -1;
  }

  bzero (mesg_out.gevent->ghello.name, 16);

  gevent_config->gconfigure.type = GrooveConfigureEvent;
  gevent_config->gconfigure.width = DisplayWidth (display, DefaultScreen(display));
  gevent_config->gconfigure.height = DisplayHeight (display, DefaultScreen(display));

  if (0 > tcp_send (gslave->tcp_conn, gevent_config)) {
    fprintf (stderr, "gslave_find_master: failed to write init event; tcp_send failed\n");
    tcp_close_connection (gslave->tcp_conn);
    gslave->tcp_conn = NULL;
    deinit_gevent (mesg_out.gevent);
    deinit_gevent (mesg_in.gevent);
    deinit_gevent (gevent_config);
    return -1;
  }

  printf ("gslave_find_master: success; connected to %s:%d\n", mesg_in.addr, mesg_in.gevent->ghelloback.port);

  deinit_gevent (mesg_out.gevent);
  deinit_gevent (mesg_in.gevent);
  deinit_gevent (gevent_config);

  return 0;
}

void
gslave_proc_motion (GSlave *gslave, GEvent *gevent)
{
  Display  *display;

  if (gslave->has_control != 1) {
    fprintf (stderr, "gslave_proc_motion: can't perform motion event; don't have control\n");
    return;
  }

  display = gslave->gdisplay.display;

  switch (gevent->gmotion.motion_type) {
    case GrooveMotionTypeAbsolute:
      XTestFakeMotionEvent (display, None, gevent->gmotion.x, gevent->gmotion.y, 0);
      break;
    case GrooveMotionTypeRelative:
      XTestFakeRelativeMotionEvent (display, gevent->gmotion.x, gevent->gmotion.y, 0);
      break;
    default:
      fprintf (stderr, "gslave_proc_motion: invalid motion type: %d\n", gevent->gmotion.motion_type);
      return;
  }

  XFlush (display);
}

void
gslave_proc_click (GSlave *gslave, GEvent *gevent)
{
  Display  *display;

  if (gslave->has_control != 1) {
    fprintf (stderr, "gslave_proc_click: can't perform click; don't have control\n");
    return;
  }

  display = gslave->gdisplay.display;

  switch (gevent->gclick.click_type) {
  case GrooveClickTypePress:
    XTestGrabControl (display, True);
    XAllowEvents (display, AsyncBoth, CurrentTime);
    XTestFakeButtonEvent (display, gevent->gclick.button, True,  0);
    XFlush (display);
    gslave->is_pressed = gevent->gclick.button;
    break;
  case GrooveClickTypeRelease:
    XTestFakeButtonEvent (display, gevent->gclick.button, False, 0);
    XTestGrabControl (display, False);
    XFlush (display);
    gslave->is_pressed = 0;
    break;
  default:
      fprintf (stderr, "gslave_proc_click: invalid click type: %d\n", gevent->gclick.click_type);
      return;
  }

  XFlush (display);
}

void
gslave_proc_keypress (GSlave *gslave, GEvent *gevent)
{
  Display  *display;

  if (gslave->has_control != 1) {
    fprintf (stderr, "gslave_proc_keypress: can't perform key press; don't have control\n");
    return;
  }

  display = gslave->gdisplay.display;

  switch (gevent->gkeypress.keypress_type) {
    case GrooveKeyPressTypePress:
      XTestFakeKeyEvent (display, gevent->gkeypress.keycode, True, 0);
      break;
    case GrooveKeyPressTypeRelease:
      XTestFakeKeyEvent (display, gevent->gkeypress.keycode, False, 0);
      break;
    default:
      fprintf (stderr, "gslave_proc_click: invalid click type: %d\n", gevent->gkeypress.keypress_type);
      return;
  }

  XFlush (display);
}

void
gslave_proc_control (GSlave *gslave, GEvent *gevent)
{
  Display  *display = gslave->gdisplay.display;
  
  switch (gevent->gcontrol.control_type) {
    case GrooveControlTypeSeize:
      printf ("gslave_proc_control: trying to seize control\n");

      if (gslave->has_control == 1) {
        fprintf (stderr, "gslave_proc_control: cant seize - I already have control\n");
        return;
      }

      gx11_pointer_pos_get (display, &gslave->prev_pos_h, &gslave->prev_pos_v);

      gslave->has_control = 1;

      break;
    case GrooveControlTypeCease:
      printf ("gslave_proc_control: trying to cease controlling\n");

      if (gslave->has_control != 1) {
        fprintf (stderr, "gslave_proc_control: can't cease - I don't have control\n");
        return;
      }

      gslave->has_control = 0;

      XConvertSelection (display, XA_PRIMARY, XA_STRING, XA_CUT_BUFFER0, gslave->gdisplay.mainwin, CurrentTime);
      XTestFakeMotionEvent (display, None, gslave->prev_pos_h, gslave->prev_pos_v, 0);
      XFlush (display);

      break;
    default:
      fprintf (stderr, "gslave_proc_control: invalid control type: %d\n", gevent->gcontrol.control_type);
      return;
  }
}

void
gslave_proc_hello (GSlave *gslave, GEvent *gevent)
{
  printf ("gslave_proc_hello: received hello event\n");
}

void
gslave_proc_helloback (GSlave *gslave, GEvent *gevent)
{
  printf ("gslave_proc_helloback: received helloback event\n");
}

void
gslave_proc_goodbye (GSlave *gslave, GEvent *gevent)
{
  printf ("gslave_proc_goodbye: received goodbye event\n");
}

void
gslave_proc_clipboard (gslave, gce)
  GSlave *gslave;
  GClipboardEvent gce;
{
  Display     *display = gslave->gdisplay.display;
  Window      mainwin = gslave->gdisplay.mainwin;

  if (gslave->clipboard_len > 0) {
    free (gslave->clipboard);
    gslave->clipboard = NULL;
    gslave->clipboard_len = 0;
  }

  gslave->clipboard = gx11_proc_clipboard (display, mainwin, gce.clipboard, gce.clipboard_len);
  gslave->clipboard_len = gce.clipboard_len;
}

void
gslave_xevent_loop (GSlave *gslave)
{
  XEvent     xevent;

  printf ("gslave_xevent_loop: start\n");

  while (1) {
    XNextEvent (gslave->gdisplay.display, &xevent);

    switch (xevent.type) {
    case SelectionClear:
      printf ("gslave_xevent_loop: SelectionClear\n");
      break;
    case SelectionNotify:
      printf ("gslave_xevent_loop: SelectionNotify\n");
      gslave_proc_selection_notify (gslave, xevent.xselection.requestor);
      break;
    case SelectionRequest:
      printf ("gslave_xevent_loop: SelectionRequest\n");
      gslave_xevent_selection_send (gslave, xevent.xselectionrequest);
      break;
    default:
      printf ("gslave_xevent_loop: %d\n", xevent.type);
      break;
    }
  }
}

void
gslave_xevent_selection_send (gslave, sre)
  GSlave *gslave;
  XSelectionRequestEvent sre;
{
  gx11_proc_selection_send (gslave->clipboard, gslave->clipboard_len, sre);
}

void
gslave_proc_selection_notify (gslave, window)
  GSlave *gslave;
  Window window;
{
  GEvent     *gevent = init_gevent ();
  Display    *display = gslave->gdisplay.display;

  if (gslave->clipboard_len > 0) {
    free (gslave->clipboard);
    gslave->clipboard = NULL;
    gslave->clipboard_len = 0;
  }

  gslave->clipboard = gx11_proc_selection_notify (display, window, &gslave->clipboard_len);

  if (NULL == (gevent->gclipboard.clipboard = malloc (gslave->clipboard_len))) {
    fprintf (stderr, "gslave_proc_selection_notify: failed in malloc()\n");
    deinit_gevent (gevent);
    return;
  }

  memcpy (gevent->gclipboard.clipboard, gslave->clipboard, gslave->clipboard_len);

  gevent->gclipboard.type = GrooveClipboardEvent;
  gevent->gclipboard.clipboard_len = gslave->clipboard_len;

  tcp_send (gslave->tcp_conn, gevent);

  deinit_gevent (gevent);
}

